########subplot#################
import numpy as np
import pandas as pd
import matplotlib.colors
import seaborn as sns
import matplotlib.pyplot as plt
from .CommonFunction import *


#    def sort_cluster(self,df,label):
#        df.sort_values(by=label,axis=1,inplace=True)
# labels:cluster numbers
# hls gist_rainbow Set1
def MetaColorMap(rawDataObj, protClustObj, path=None, selectFeatureSer=None, clustColor='Set1', figSizeLi=[10, 5],
                 xlabelSizeStr='large', ylabelSizeStr='small', ylabelDist=50, dpi=96):
    '''
    draw a figure about different feature value in meta data

    :param rawDataObj: ds_inputRawData object
    :param protClustObj: ds_protClust object
    :param path: the directory to save the figure, if none, the figure will only show in the program
    :param clustColor: the color palette used in the figure, default is Set1 in matplotlib
    :param figSizeLi: the width and high of the figure, default is 10 and 5
    :param xlabelSizeStr: x label font size
    :param ylabelSizeStr: y label font size
    :param ylabelDist: y label distance
    :return:
    '''

    metaDataDf=rawDataObj.GetMetadataDf ().copy()
    if selectFeatureSer is None:
        labelSer = rawDataObj.GetMetadataDf ().columns.values
    elif type(selectFeatureSer) is str:

        labelSer = pd.read_csv(selectFeatureSer, header=None, index_col=False)
        labelSer = labelSer.transpose()[0]

    else :
        labelSer=selectFeatureSer

    metaDataDf = metaDataDf[labelSer]
    # sort metadata by cluster

    clustNumSer = protClustObj.clustResultObj.sampleClustSer
    metaDataDf.insert(0,'clusterList', clustNumSer)

    ###sort same cluster together###
    metaDataDf.sort_values (by='clusterList', axis=0, inplace=True)
    countCol = metaDataDf.shape[1]  # how many columns in a row
    lenEachClustLi = GetlenEachClustLi (protClustObj)
    #metaDataDf = metaDataDf.drop (columns=['clusterNumList'])







    fig, axes = plt.subplots (countCol, 1, figsize=(figSizeLi[0], figSizeLi[1]), gridspec_kw={'hspace': 0.01, 'wspace': 0})

    for rowLabel in range (countCol):
        currentColSer = metaDataDf.iloc[:, rowLabel].copy()
        featureValueNum = len (set (currentColSer))
        currentColSer = currentColSer.astype ('str')
        #currentColSer = np.ma.masked_where (currentColSer == 'NA', currentColSer, False)
        # don't know how to name  1: Stage IV,stage III...      2: 6,6,7,7,3,4,1

        labelContentAr, contentOrderAr = np.unique (currentColSer, return_inverse=True)
        currentColSer = __ConvertToStr (labelContentAr, currentColSer)

        currentColSer =np.ma.array(currentColSer,mask=np.isnan(currentColSer))
        # use contentOrderAr draw color bar

        barContentAr = currentColSer.reshape (currentColSer.shape)

        pltColors = sns.color_palette (clustColor, featureValueNum)

        # add dimension of barContentAr. axis=0 means row; axis=1 means column
        barContentAr = np.expand_dims (barContentAr, axis=0)
        # map the color and the content order
        cmap = matplotlib.colors.ListedColormap (pltColors)
        # set mask color
        cmap.set_bad('white')
        norm = matplotlib.colors.BoundaryNorm (np.arange (featureValueNum + 1) - 0.5, featureValueNum)

        # remove xlabel, yticklabels, xticklabels; set ylabel as row label

        axes[rowLabel].set_ylabel (metaDataDf.columns.values[rowLabel], rotation=0, horizontalalignment='center', size=ylabelSizeStr,
                                   labelpad=ylabelDist)

        axes[rowLabel].set_yticks ([])
        axes[rowLabel].set_xticks ([])

        axes[rowLabel].imshow (barContentAr, cmap=cmap, norm=norm, aspect='auto')


        # x-axis,y-axis,width,length
    axes[countCol-1 ].set_xlabel ('Samples', position=(0.98, 0.1, 3, 10), size=xlabelSizeStr)

    if path==None:
        plt.show()
    else :
        fig.savefig (path,bbox_inches="tight", dpi=dpi)


def __ConvertToStr(labelContentAr, currentColSer):
    resultSer = currentColSer.copy ()
    #labelContentAr = labelContentAr[~labelContentAr.mask]
    #print(currentColSer[~currentColSer.mask])
    intNum = 0.0


    for label in labelContentAr:

        for data in range (len (currentColSer)):
            if currentColSer[data] == 'NA':
                resultSer[data] = float ('nan')
            elif label == currentColSer[data]:
                resultSer[data] = float(intNum)

        intNum += 1
    resultSer=resultSer.astype('float')
    return resultSer

def __xtickLen(lenEachClustLi):
    result=[]
    sum=0
    for k in lenEachClustLi:
        sum = sum + k/2
        result.append(sum)
        sum=sum+k/2
    return result





from .CommonFunction import *


def CalcFeatureMaxFscore(rawDataObj, protClustObj):
    """
    傳入inputRawData和protClust及clustering結果，計算各個metadata feauture的Fscore並儲存至ds_protClust object中的featureMaxFscoreDi dictionary

    :param rawDataObj: ds_inputRawData object
    :param protClustObj: ds_protClust object
    :param predLi: clutering result list，大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :return: ds_protClust object
    """
    metaColNameLi = rawDataObj.GetMetadataDf().columns.values
    metaDf = rawDataObj.GetMetadataDf().copy()
    predLi = protClustObj.clustResultObj.sampleClustSer.copy()
    #predLi=list(TurnClustToNumSer(rawDataObj.GetMetadataDf().index.values, protClustObj))
    protClustObj.clustScoringObj.featureMaxFscoreDi = dict()
    for colNameStr in metaColNameLi:
        trueLi = TurnLabelToNum(metaDf[colNameStr])
        predLi = TurnLabelToNum(predLi)
        maxFscore=GetMaxFscore(trueLi, predLi)
        protClustObj.clustScoringObj.featureMaxFscoreDi.update({colNameStr: maxFscore})

    return protClustObj

# 傳入clustering 結果 & class label (both are list)，計算Fscore
# trueLi --> 大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
def GetMaxFscore(trueLi, predLi):

    """
    傳入clustering 結果 & class label (both are list)，計算Fscore

    :param trueLi: feature class list,大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param predLi: clutering result list，大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :return: float Fscore
    """
    finalMaxFscore=0

    for predNum in range(len(set(predLi))):

        maxFscore=0


        for trueNum in range(len(set(trueLi))):
            recall=_ComputeRecall(trueLi, predLi, trueNum, predNum)
            precision=_ComputePrecision(trueLi, predLi, trueNum, predNum)

            if recall==0 and precision==0:
                localFscore=0
            else:
                localFscore=2*(precision*recall)/( precision+recall)

            if localFscore>=maxFscore :
                maxFscore=localFscore

        truecount = _CountTrueNum(predLi, predNum)
        finalMaxFscore+=(maxFscore * ( truecount/ len(predLi)))

    return finalMaxFscore


def _ComputeRecall(trueLi, predLi, trueNum, predNum):
    """
    compute recall for computing Fscore   TP/(TP+FN)

    :param trueLi: feature class list,大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param predLi: clutering result list，大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param trueNum: int, represent current class
    :param predNum: int, represent current cluster
    :return: float recall
    """
    truePositive=0
    falseNagetive=0
    for lens in range(len(trueLi)):
        if predLi[lens]==predNum and trueLi[lens]==trueNum:
            truePositive+=1
        elif trueLi[lens]==trueNum:
            falseNagetive+=1

    return truePositive/(truePositive+falseNagetive)


def _ComputePrecision(trueLi, predLi, trueNum, predNum):
    """
    compute precision for computing Fscore   TP/(TP+FP)

    :param trueLi: feature class list,大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param predLi: clutering result list，大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param trueNum: int, represent current class
    :param predNum: int, represent current cluster
    :return: float precision
    """
    truePosive = 0
    falsePositive = 0
    for lens in range(len(trueLi)):
        if predLi[lens] == predNum and trueLi[lens] == trueNum:
            truePosive += 1
        elif predLi[lens] == predNum:
            falsePositive += 1

    return truePosive / (truePosive + falsePositive)


def _CountTrueNum(trueLi, trueNum):
    """
    count the number of current class

    :param trueLi: feature class list,大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param trueNum: int, represent current class
    :return: int, the number of current class
    """
    count=0
    for lens in range(len(trueLi)):
        if trueLi[lens]==trueNum:
            count+=1
    return count


from sklearn import cluster, metrics, datasets
from nltk.cluster import KMeansClusterer, euclidean_distance
from .CommonFunction import *
import numpy as np


def KmeansClust_sklearn(rawDataObj, protClustObj):
    """
    K means clustering 

    :param rawDataObj: ds_InputRawData
    :type rawDataObj: object 
    :param protClustObj: ds_ProtClust
    :type protClustObj: object
    :returns: predArr 

    .. note::
        The parameter in protClustObj.clustParamObj.__kwargs must include 'clustNum', 'seed' \n
        distance measure only euclidean distance
    """
    k = protClustObj.clustParamObj.GetParam('clustNum')
    seed = protClustObj.clustParamObj.GetParam('seed')
    dataArr = rawDataObj.GetProtRatioDf().values

    kmeans_fit = cluster.KMeans(n_clusters=k, random_state=seed).fit(dataArr)
    predArr = kmeans_fit.labels_

    return predArr


def KmeansClust_nltk(rawDataObj, protClustObj):
    """
    Nltk K means clustering 

    :param rawDataObj: ds_InputRawData
    :type rawDataObj: object 
    :param protClustObj: ds_ProtClust
    :type protClustObj: object
    :returns:   predArr 

    .. note::
        The parameter in protClustObj.clustParamObj.__kwargs must include 'clustNum', 'distance'\n
        accepted distance : 'euclidean', 'manhattan', 'cosine', 'pcc';
    """
    k = protClustObj.clustParamObj.GetParam('clustNum')
    dist = protClustObj.clustParamObj.GetParam('kmeansDisatance')
    distance = GetDistance(dist)
    dataArr = rawDataObj.GetProtRatioDf().values

    clusterer = KMeansClusterer(num_means=k, distance=distance, repeats=1)
    predArr = np.array(clusterer.cluster(dataArr, True))

    return predArr




from .ReadFiles import ReadFilesPath, CheckRatioNotEmpty, CheckMetadataNotEmpty
from .Filter import FilterByRank, FilterByCutoff, FilterByTopPercentage, FilterByGeneList
from .Nmlz import  ZscoreNmlz
from .Log import LogRatioToRatio, RatioToLogRatio
from .DataCleaning import RemoveProtWithNa
from .Imputation import ImputeMin, ImputeAverage, ImputeKnn
from .SortingProtRatioDf import SortByColumn, SortByRow
from .Serialization import *
__all__ =["ReadFilesPath", "CheckRatioNotEmpty", "CheckMetadataNotEmpty",
          "FilterByRank", "FilterByCutoff", "FilterByTopPercentage", "FilterByGeneList",
          "ZscoreNmlz",
          "LogRatioToRatio", "RatioToLogRatio",
          "RemoveProtWithNa",
          "ImputeMin", "ImputeAverage", "ImputeKnn",
          "SortByRow", "SortByColumn"]



import math
from collections import Counter
import pandas as pd
import numpy as np 
import scipy.spatial as spatial #for partition index and seperation index


def ConvertPredLiStartZero(predLi):
    '''
    predLi is sampleClustSer ([1, 2, 2, 3, 3, 4], sometimes for the convenience need to convert it to [0, 1, 1, 2, 2 ,3]

    :param predLi: sampleClustSer's list [1, 2, 2, 3, 3, 4]
    :return: list [0, 1, 1, 2, 2 ,3]
    '''
    resultLi=[]
    predRangeDict = {}
    count = 0
    for i in set(predLi):
        predRangeDict[i] = count
        count+=1

    for i in range(len(predLi)):
        resultLi.append(predRangeDict[predLi[i]])

    return resultLi


#for partition index and seperation index
def GetClusterCenter(dataArr, predLi, num):
    '''
    Get the center of each cluster.

    :param dataArr: np.array, data values, shape [data_num, feature_num]
    :param: predLi: list, dimension [data_num, ], the number in preds starts from 0, no integer is skipped
    :param: num: int, number of clusters
    :return: center: np.array, center of the clusters, dimension [cluster_num, feature_num] 
    '''
    centerArr = np.zeros((num,dataArr.shape[1]))
    clusMemCountArr = np.zeros(num)
    predFromZeroLi = ConvertPredLiStartZero(predLi)
    for i in range(len(predFromZeroLi)):
        centerArr[predFromZeroLi[i]] += dataArr[i]
        clusMemCountArr[predFromZeroLi[i]] += 1

    centerArr = np.divide(centerArr.T,clusMemCountArr).T
    return centerArr

#for partition index and seperation index
def GetUtilityMatrix(dataArr, centerArr, distanceMetric):
    '''
    Get the utility Matrix for the clustering results.
    The membership value is inversely proportional to distance (distance from the data point to each center) square.

    :param: dataArr: np.array. Shape [data_num feature_num]
    :param: center: np.array. Center of the clusters, dimension [cluster_num, feature_num]
    :param: distanceMetric: string. The distance metric. Possible values: euclidean, cityblock, cosine, correlation
    :return: np.array. Shape [cluster_num, data_num]
    '''
    num = centerArr.shape[0]

    distanceMatrixArr = spatial.distance.cdist(dataArr, centerArr, metric=distanceMetric)
    distanceMatrixArr = distanceMatrixArr/(np.sum(distanceMatrixArr,axis=1).reshape(-1,1))
   
    with np.errstate(divide='ignore', invalid='ignore'):
        utilityArr = 1/np.square(distanceMatrixArr)
        utilityArr[distanceMatrixArr == 0] = 1

    utilityArr = utilityArr/(np.sum(utilityArr,axis=1).reshape(-1,1)) #[data_num, cluster_num]

    return utilityArr.T

def ComputeEntropy(featureLi, unit='shannon'):
    '''
    compute entropy of data(meta data feature)

    :param featureLi: label list or number list, len is same as sample numbers
    :param unit: the log base of computing entropy. shannon=2, natural=e, hartley=10
    :return: float entropy
    '''



    base = {
        'shannon': 2.,
        'natural': math.exp(1),
        'hartley': 10.
    }

    if len(featureLi) <= 1:
        return 0

    counts = Counter()

    for d in featureLi:
        counts[d] += 1

    ent = 0

    probs = [float(c) / len(featureLi) for c in counts.values()]
    for p in probs:
        if p > 0.:
            ent -= p * math.log(p, base[unit])

    return ent



def TurnLabelToNum(featureLi):
    '''
    change feature value list into number list, in order to compute scores ex: ['black','black','red','yellow'] -> [0,0,1,2]

    :param featureLi: feature value list
    :return: number list
    '''
    listNum=0
    for label in set(featureLi):

        for i in range(len(featureLi)):
            if featureLi[i]==label:
                featureLi[i]=listNum
        listNum=listNum+1


    return list(featureLi)



def TurnClustToNumSer(sampleNameLi, protClustObj):
    '''
    change clustSampleSer in clustDataObj var into a numlist for the following figure

    :param sampleNameLi: the name list that needed to transfer
    :param protClustObj: ds_protClust object
    :return: numlist
    '''

    clustSampleSer= protClustObj.clustResultObj.clustSampleSer

    numSer = pd.Series([0 for x in range(len(sampleNameLi))],index=sampleNameLi)

    for col in range(len(clustSampleSer)):
        for row in range(len(clustSampleSer[col])):

            for num in range(len(sampleNameLi)):
                if clustSampleSer[col][row] == sampleNameLi[num]:
                    numSer[num] = col
    return numSer



def GetlenEachClustLi(protDataObj):
    '''
    count length of each cluster in protDataObj.clustSampleSer

    :param protDataObj:
    :return: lentgth list. Example:[7,9,7] for 3 clusters
    '''

    clustSampleLi = list(protDataObj.clustResultObj.clustSampleSer)
    lenLi = [0 for x in range(len(clustSampleLi))]
    for col in range(len(clustSampleLi)):
        lenLi[col]=len(clustSampleLi[col])

    return lenLi




from statsmodels.graphics.mosaicplot import mosaic
import matplotlib.colors
from pylab import *
import numpy as np
import pandas as pd




def DrawMosaicPlot(rawDataObj, protClustObj, FeatureStr, titleStr='Mosaic Plot', colorLi=None, figSizeLi=[10, 10],
                path=None, dpi=96, legenCol=6, legendLoc=[0.5, -0.1]):
    """
    Draw Mosaic plot to show the number of feature in each cluster

    :param rawDataObj: {ds_InputRawData }
    :param protClustObj: {ds_ProtClust }
    :param FeatureStr: the metadata feature you want to select (String)
    :param titleStr: the title of the plot (String)
    :param colorLi: the string of color set or the marker color, list of Hex color codes, default is 'paired'
    :param figSizeLi: the size of the plot (String)
    :param path: If is null then show the plot. Else store to that path.
    :param dpi: Resolution of the figure. (default = 96)
    :param legenCol: the number of value each column in legend
    :param legendLoc: the location of legend
    :return: figure
    """

    clustNumSer = protClustObj.clustResultObj.sampleClustSer.copy()

    metadataDf = rawDataObj.GetMetadataDf().copy()
    dataDf = pd.DataFrame({'Cluster':clustNumSer,FeatureStr:metadataDf[FeatureStr]})
    dataDf = dataDf.sort_values('Cluster')
    labelDict ={}
    colorDict ={}
    ### produce Dictionary tuple list to delete the content in the map
    for x in sorted(set(clustNumSer)):
        for y in sorted(set(metadataDf[FeatureStr])):
            labelDict.update({(str(x),y):''})


    ## colormap
    if colorLi==None or isinstance(colorLi,str):
        if isinstance(colorLi,str):
            colorSetStr = colorLi
        else:
            colorSetStr = 'Paired'
        colorLi=[]
        cmap = cm.get_cmap(colorSetStr, len(set(metadataDf[FeatureStr])))  # PiYG

        for i in range(cmap.N):
            rgb = cmap(i)[:3]  # will return rgba, we take only first 3 so we get rgb
            colorLi.append(matplotlib.colors.rgb2hex(rgb))


    colorDict = dict(zip(sorted(set(metadataDf[FeatureStr])), colorLi))
    print(colorDict)
    for x in sorted(set(clustNumSer)):
        for y in sorted(set(metadataDf[FeatureStr])):
            colorDict.update({(str(x),y):colorDict[y]})

    fig, axes = plt.subplots(1,1, figsize=(figSizeLi[0], figSizeLi[1]), dpi=dpi)

    #props = lambda key: {'color': '#ff7f0e' if '1' in key else ('gray' if 'luad' in key else 'blue')}
    props = lambda key:{'color':colorDict[key]}
    labelizer = lambda k: labelDict[k]
    a,b = mosaic(dataDf,['Cluster',FeatureStr], title=titleStr, properties=props, labelizer=labelizer, ax=axes, axes_label=True, gap=0.05)
    axes.set_yticklabels([])
    #print(dataDf[FeatureStr])

    leg = axes.legend(_BuildColorList(dataDf[FeatureStr].tolist()),loc='center', bbox_to_anchor=(legendLoc[0], legendLoc[1]),ncol=legenCol)

    #plt.xlabel('Cluster')

    if path is None:
        plt.show()
    else:
        plt.savefig('{}.png'.format(path), bbox_inches='tight')
        plt.close()

    return fig




def _BuildColorList(SearchLi):
    colLi = []
    for i in SearchLi:
        if i in colLi:
            continue
        else:
            colLi.append(i)
    return colLi


def PrintParagraph(rawDataObj, protClustObj):
    """
    print some paragraph about the clustering result

    :param rawDataObj:
    :type rawDataObj: object
    :param protClustObj:
    :type  protClustObj: object
    """
    clustSer = protClustObj.clustResultObj.clustSampleSer.values
    clustNum = protClustObj.clustResultObj.clustNum

    print("\nClustering method \n")

    paramLi = list(protClustObj.clustParamObj.GetParamLi())
    methodStr = protClustObj.clustParamObj.GetParam('method')
    k = protClustObj.clustParamObj.GetParam('clustNum')
    print("\tThe clustering method used in this run is \"{}\".\n\tThe cluster number is \"{}\".".format(methodStr,clustNum))
    print("\tOther parameters used in this run :")
    for i in paramLi:
        if i not in ['method','clustNum']:
            print("\t\t{} : {}".format(i,protClustObj.clustParamObj.GetParam(i)))
    sampleStr = ""
    for i in range(1,clustNum+1):
        sampleStr += str(len(clustSer[i-1]))
        if i == clustNum-1:
            sampleStr += ' and '
        else:
            sampleStr += ', '

    print("\nClustering result \n")

    print("\tThe samples are clustered into {} clusters .".format(clustNum))
    print("\tThe number of sample in cluster1 to cluster{} are {}respectively.".format(clustNum,sampleStr))
    
    print("\nClustering scores \n")
    # print("\tpartitionIndex : {} , separationIndex : {} , silhScore : {} , purity : {} , giniScore : {}\n" \
                    # .format(protClustObj.clustScoringObj.partitionIndex,protClustObj.clustScoringObj.separationIndex,\
                        # protClustObj.clustScoringObj.silhScore,protClustObj.clustScoringObj.purity,protClustObj.clustScoringObj.giniScore))
    print("\tpartitionIndex : ",protClustObj.clustScoringObj.partitionIndex)
    print("\tseparationIndex : ",protClustObj.clustScoringObj.separationIndex)
    print("\tsilhScore : ",protClustObj.clustScoringObj.silhScore)
    print("\tpurity : ",protClustObj.clustScoringObj.purity)
    print("\tginiScore : ",protClustObj.clustScoringObj.giniScore)
        # print("There are {} samples in {}th cluster :\n {}\n".format(len(clustSer[i-1]),i,clustSer[i-1]))
    # print("Utility Matrix Data frame :\n{}".format(protClustObj.clustResultObj.utilityMatrixDf))
    # print("Sample Series for each cluster :\n{}".format(protClustObj.clustResultObj.clustSampleSer))
    # print("Cluster ID for each sample :\n{}".format(protClustObj.clustResultObj.sampleClustSer))




import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import numpy as np
import os.path


def FindGoTermOverlap(upOrDownStr="up", processStr="biological", clusterNum=5, goTerm1Str="gene ontology",
                      goTerm2Str="gene ontology_consensus"):
    """
    find two clustering results' go terms overlap

    :param upOrDownStr: is set in file name, means the genes chose from rank sum test each subtype are based on > or < than rest subtypes. (although we not use down anymore)
    :param processStr: "biological", "cellular", or "molecular"
    :param clusterNum: the number of cluster
    :param goTerm1Str: go terms file name
    :param goTerm2Str: go terms file name

    :return: heatmap ratioDf

    """
    ratioDf = pd.DataFrame([[0.0 for x in range(clusterNum)]] * clusterNum)

    for col in range(clusterNum):
        path1Str = "Example/" + goTerm1Str + "/cluster" + str(col + 1) + "_" + upOrDownStr + "/" + processStr + "/cluster" + str(
            col + 1) + upOrDownStr + ".txt"

        if os.path.isfile(path1Str):
            goTerm1Df = pd.read_csv(path1Str, delimiter="\t")
        else:
            continue
        for row in range(clusterNum):
            path2Str = "Example/" + goTerm2Str + "/cluster" + str(
                row + 1) + "_" + upOrDownStr + "/" + processStr + "/cluster" + str(
                row + 1) + upOrDownStr + ".txt"
            if os.path.isfile(path2Str):
                goTerm2Df = pd.read_csv(path2Str, delimiter="\t")
            else:
                continue
            count = 0
            num = 0
            for goNameStr in goTerm1Df['geneSet']:
                num += 1

                for goName2Str in goTerm2Df['geneSet']:
                    if goNameStr == goName2Str:
                        count += 1
            totalGoTerm = _CountTotalGoTerm(goTerm1Df['geneSet'], goTerm2Df['geneSet'])
            ratioDf[col][row] = count / totalGoTerm
    print(ratioDf)

    __PrintgoTermNum(clusterNum, goTerm1Str, goTerm2Str, upOrDownStr=upOrDownStr, processStr=processStr)
    sns.heatmap(ratioDf, cmap="Reds", linewidths=.5, annot=True, vmin=0, vmax=1,
                xticklabels=[x + 1 for x in range(clusterNum)], yticklabels=[x + 1 for x in range(clusterNum)])
    plt.xlabel('ConKM_eu')
    plt.ylabel('HC_cos_comp')
    plt.savefig('Con_HC_0.4heatmap.png',dpi=300)

    return ratioDf


def _CountTotalGoTerm(goTerm1Ser, goTerm2Ser):
    """
    count the total number of two go terms without repeat
    :param goTerm1Ser: the gene set ID series of go term1. ex:[Go1,Go2,Go3]
    :param goTerm2Ser:the gene set ID series of go term2. ex:[Go1,Go2,Go3]
    :return: int, the number of two go terms
    """
    totalLi = goTerm1Ser.tolist()
    for name in goTerm2Ser:
        if _IsgoNameInOverlapLi(totalLi, name):
            continue
        else:
            totalLi.append(name)
    return len(totalLi)


def _IsgoNameInOverlapLi(geneSetLi, goName):
    """
    to judge if the gene set ID is in the go term list
    :param geneSetLi:a list contain gene set IDs
    :param goName: gene set ID
    :return: True/False
    """
    if geneSetLi == []:
        return False
    for name in geneSetLi:
        if name == goName:
            return True
    return False


def _FindGoTermOverlapLi(goTerm1Str, goTerm2Str, choseGoNum=50, upOrDownStr="up", processStr="biological",
                         clusterNum=5):
    """
    find the overlap Gene set list between (param: goTerm1Str) and (param: goTerm2Str),
    for each subtype we choose top (param: choseGoNum) overlap gene sets to the overlap list
    :param choseGoNum: the number of overlap gene sets we choose each subtype
    :param goTerm1Str: go terms file name
    :param goTerm2Str: go terms file name
    :param upOrDownStr: is set in file name, means the genes chose from rank sum test each subtype are based on > or <
    than rest subtypes. (although we not use down anymore)
    :param processStr: "biological", "cellular", or "molecular"
    :param clusterNum: the number of cluster
    :return: overlap list
    """
    overlapLi = []
    for col in range(clusterNum):
        path1Str = "Example/" + goTerm1Str + "/cluster" + str(col + 1) + "_" + upOrDownStr + "/" + processStr + "/cluster" + str(
            col + 1) + upOrDownStr + ".txt"

        if os.path.isfile(path1Str):
            goTerm1Df = pd.read_csv(path1Str, delimiter="\t")
        else:

            continue

        count = 0

        for goNameStr in goTerm1Df['geneSet']:
            if count >= choseGoNum:
                break
            elif _IsgoNameInOverlapLi(overlapLi, goNameStr):
                count += 1
                continue
            else:
                overlapLi.append(goNameStr)
                count += 1

    for col in range(clusterNum):
        path2Str = "Example/" + goTerm2Str + "/cluster" + str(
            col + 1) + "_" + upOrDownStr + "/" + processStr + "/cluster" + str(
            col + 1) + upOrDownStr + ".txt"
        if os.path.isfile(path2Str):
            goTerm2Df = pd.read_csv(path2Str, delimiter="\t")
        else:

            continue

        count = 0
        for goName2 in goTerm2Df['geneSet']:
            if count >= choseGoNum:
                break
            elif _IsgoNameInOverlapLi(overlapLi, goName2):
                count += 1
                continue
            else:
                overlapLi.append(goName2)
                count += 1

    print(overlapLi)

    print(len(overlapLi))
    return overlapLi


def _FindGoTermPValue(overlapLi, goTermClustLi, goTermStr, upOrDownStr="up", processStr="biological",
                      clusterNum=5):
    """
    base on overlapLi to find the gene sets' pvalue, in order to draw gene sets heatmap next step
    :param overlapLi: overlap Gene set list between (param: goTerm1Str) and (param: goTerm2Str)
    :param goTermClustLi: the correspond subtype order of the goTermStr, based on goTerm overlap heatmap.
    ex: two goTermStrs  have 5 subtype and goTermStr1_subtype1 similar to goTermStr2_subtype3, and the goTermClustLi
    of goTermStr1 is [1, 2, 3, 4, 5], than the goTermClustLi of goTermStr2 may be [3, 2, 1, 4 ,5]
    :param goTermStr: go terms file name
    :param upOrDownStr: is set in file name, means the genes chose from rank sum test each subtype are based on > or <
    than rest subtypes. (although we not use down anymore)
    :param processStr: "biological", "cellular", or "molecular"
    :param clusterNum: the number of cluster
    :return: pValueDf
    """
    pValueDf = pd.DataFrame([[1.0 for x in range(len(overlapLi))]] * clusterNum, columns=overlapLi)

    count = 0
    for runingClustering in range(clusterNum):
        pathStr = "Example/" + goTermStr + "/cluster" + str(
            runingClustering + 1) + "_" + upOrDownStr + "/" + processStr + "/cluster" \
               + str(runingClustering + 1) + upOrDownStr + ".txt"
        if os.path.isfile(pathStr):
            goTermDf = pd.read_csv(pathStr, delimiter="\t")
        else:
            count += 1
            continue

        goTerm2Num = goTermDf.shape[0]

        for nameStr in overlapLi:
            for i in range(goTerm2Num):

                if nameStr == goTermDf['geneSet'][i]:

                    goTermNameStr = goTermDf['geneSet'][i]
                    if goTermDf['pValue'][i] == 0:
                        pValueDf[goTermNameStr][_FindCorrespondRow(goTermClustLi, count)] = 1E-17

                    else:
                        pValueDf[goTermNameStr][_FindCorrespondRow(goTermClustLi, count)] = float(goTermDf['pValue'][i])
                        # print(goTerm['pValue'][i])
        count += 1
    return pValueDf


def DrawGoTermHeatMap(goTerm1Str, goTerm2Str, goTerm1ClustLi=None, goTerm2ClustLi=None, choseGoNum=50,
                      upOrDownStr="up", processStr="biological", clusterNum=5, dpi=96):
    """
    :param upOrDownStr: is set in file name, means the genes chose from rank sum test each subtype are based on > or <
    than rest subtypes. (although we not use down anymore)
    :param processStr: "biological", "cellular", or "molecular"
    :param clusterNum: the number of cluster
    :param goTerm1Str: go terms file name
    :param goTerm2Str: go terms file name
    :param goTerm1ClustLi:the correspond subtype order of the goTermStr, based on goTerm overlap heatmap.
    ex: two goTermStrs  have 5 subtype and goTermStr1_subtype1 similar to goTermStr2_subtype3, and the goTermClustLi
    of goTermStr1 is [1, 2, 3, 4, 5], than the goTermClustLi of goTermStr2 may be [3, 2, 1, 4 ,5]
    :param goTerm2ClustLi: same as above
    :param choseGoNum:  the number of overlap gene sets we choose each subtype
    :return:
    """
    if goTerm1ClustLi is None:
        goTerm1ClustLi = [x+1 for x in range(clusterNum)]
    if goTerm2ClustLi is None:
        goTerm2ClustLi = [x+1 for x in range(clusterNum)]
    overlapLi = _FindGoTermOverlapLi(choseGoNum=choseGoNum, upOrDownStr=upOrDownStr,
                                     processStr=processStr, clusterNum=clusterNum,
                                     goTerm1Str=goTerm1Str, goTerm2Str=goTerm2Str)

    goTermResultDf = _FindGoTermPValue(overlapLi, goTerm2ClustLi, goTermStr=goTerm2Str, upOrDownStr=upOrDownStr,
                                       processStr=processStr,
                                       clusterNum=clusterNum)
    goTermResultDf.to_csv(
        "Example/" + goTerm2Str + "/pValueTable_" + str(choseGoNum) + "_" + upOrDownStr + "_" + processStr + ".csv")

    goTermResultDf = np.log10(goTermResultDf)
    # clors=sns.palplot(sns.light_palette("red", reverse=True))
    fig, axes = plt.subplots(1, 1, figsize=(30, 3))
    mask = goTermResultDf == 0
    # yltick=__Setytick(goTerm1Clust, goTerm2Clust), xticklabels=goTermResultDf.columns
    sns.heatmap(goTermResultDf, cmap="Reds_r", mask=mask, linewidths=.5, ax=axes, xticklabels=False,
                yticklabels=goTerm2ClustLi)

    fig.savefig(
        "Example/" + goTerm2Str + "/pValueTable_" + str(choseGoNum) + "_" + upOrDownStr + "_" + processStr + ".png",
        bbox_inches="tight", dpi=dpi)

    goTermResultDf = _FindGoTermPValue(overlapLi, goTerm1ClustLi, goTermStr=goTerm1Str, upOrDownStr=upOrDownStr,
                                       processStr=processStr,
                                       clusterNum=clusterNum)
    goTermResultDf.to_csv(
        "Example/" + goTerm1Str + "/pValueTable_" + str(choseGoNum) + "_" + upOrDownStr + "_" + processStr + ".csv")

    goTermResultDf = np.log10(goTermResultDf)
    # clors=sns.palplot(sns.light_palette("red", reverse=True))
    fig, axes = plt.subplots(1, 1, figsize=(30, 3))
    mask = goTermResultDf == 0
    # RdYlBu
    sns.heatmap(goTermResultDf, cmap="Reds_r", mask=mask, ax=axes,xticklabels=False,
                yticklabels=goTerm1ClustLi)

    fig.savefig(
        "Example/" + goTerm1Str + "/pValueTable_" + str(choseGoNum) + "_" + upOrDownStr + "_" + processStr + ".png",
        bbox_inches="tight",dpi=dpi)




def __PrintgoTermNum(clusterNum, goTerm1Str, goTerm2Str, upOrDownStr="up", processStr="biological"):
    '''
    Print how many goterms in each cluster subtypes
    :param upOrDownStr: is set in file name, means the genes chose from rank sum test each subtype are based on > or <
    than rest subtypes. (although we not use down anymore)
    :param processStr: "biological", "cellular", or "molecular"
    :param clusterNum: the number of cluster
    :param goTerm1Str: go terms file name
    :param goTerm2Str: go terms file name
    :return: None
    '''
    print("Method " + goTerm1Str)
    for num in range(clusterNum):
        path1 = "Example/" + goTerm1Str + "/cluster" + str(num + 1) + "_" + upOrDownStr + "/" + processStr + "/cluster" + str(
            num + 1) + upOrDownStr + ".txt"
        if os.path.isfile(path1):
            goTerm1 = pd.read_csv(path1, delimiter="\t")
        else:
            print("cluster" + str(num + 1) + " has no gene set")
            continue
        print("cluster" + str(num + 1) + ": " + str(len(goTerm1['geneSet'])))

    print("Method " + goTerm2Str)
    for num in range(clusterNum):
        path2 = "Example/" + goTerm2Str + "/cluster" + str(num + 1) + "_" + upOrDownStr + "/" + processStr + "/cluster" + str(
            num + 1) + upOrDownStr + ".txt"
        if os.path.isfile(path2):
            goTerm2 = pd.read_csv(path2, delimiter="\t")
        else:
            print("cluster" + str(num + 1) + " has no gene set")
            continue
        print("cluster" + str(num + 1) + ": " + str(len(goTerm2['geneSet'])))

def _FindCorrespondRow(goTermClustLi, count):
    """
    put correspond goterm row in to right row
    :param goTermClustLi: the correspond subtype order of the goTermStr, based on goTerm overlap heatmap.
    ex: two goTermStrs  have 5 subtype and goTermStr1_subtype1 similar to goTermStr2_subtype3, and the goTermClustLi
    of goTermStr1 is [1, 2, 3, 4, 5], than the goTermClustLi of goTermStr2 may be [3, 2, 1, 4 ,5]
    :param count:int, current gene set position in gene set list
    :return: int, the row in gene sets heatmap
    """
    result = 0
    for go in goTermClustLi:
        if go == count + 1:
            return result

        result += 1


"""
Contain ReadFilesPath(), CheckRatioNotEmpty(), CheckMetadataNotEmpty
"""

import pandas as pd
import numpy as np
from DataStructure import ds_InputRawData
from .CommonFunction import GenerateStat, PrintStat
from .DataChecking import CheckFileExistence, CheckPatientNum
from .DataCleaning import RemoveContinuousFeature, AssignNaStrToNotAvailableCell, RemoveProtWithNa


def ReadFilesPath(protRatioPath, metadataPath, metadataCategoryPath=''):
    """
   
    :param protRatioPath: string of path of protein ratio file (protein and patient as row and column)
    :type protRatioPath: str
    :param metadataPath: string of path of metadata file (patient and metadata as row and column)
    :type metadataPath: str
    :param metadataCategoryPath: string of path of file that determines metadata are categorical or not.

        file content: row1 is feature labels, row2 is 0 (continuous) or 1 (categorical).
        ::
            +---------------+----------+----------+-----+
            | name          | feature1 | feature2 | ... /
            +============+============+===========+=====+
            | isCategorical | 0        | 1        | ... /
            +---------------+----------+----------+-----+

    :return: ds_InputRawData
    :rtype: object
    """

    # Input
    protRatioDf = pd.read_csv(protRatioPath, index_col=0)
    protRatioDf_transposed = protRatioDf.T  # Transpose protRatioDf (patient and protein as row and column)
    metadataDf = pd.read_csv(metadataPath, index_col=0)

    # Check path cannot be ""
    if len(metadataCategoryPath) != 0:
        metadataCategoryDf = pd.read_csv(metadataCategoryPath, index_col=0)
    else:
        metadataCategoryDf = pd.DataFrame()
    # Check pd.DataFrame() cannot be empty
    if not(CheckFileExistence(protRatioDf=protRatioDf_transposed, metadataDf=metadataDf)):
        return 0
    # Check number of patients is equal in protein ratio and metadata
    if not(CheckPatientNum(protRatioDf=protRatioDf_transposed, metadataDf=metadataDf)):
        return 0

    # Make lowercase
    metadataDf = metadataDf.applymap(lambda x: x.lower() if type(x) == str else x)
    # Drop continuous columns of metadataDf
    metadataDf = RemoveContinuousFeature(metadataDf=metadataDf, metadataCategoryDf=metadataCategoryDf)
    # Deal with metadata's missing value
    metadataDf = AssignNaStrToNotAvailableCell(metadataDf)

    # Create ds_InputRawData object
    rawDataObj = ds_InputRawData(protRatioDf_transposed, metadataDf)
    rawDataObj.patientNameLi = list(protRatioDf_transposed.columns[:])
    rawDataObj = GenerateStat(rawDataObj)

    if protRatioDf_transposed.isnull().values.any():
        print("Need call RemoveProtWithNa() or Impute later.\n")

    return rawDataObj


def CheckRatioNotEmpty(ds_InputRawDataObj):
    """
    Check protRatio is not empty.

    :param ds_InputRawDataObj:
    :type ds_InputRawData: object
    :return: True if protRatioDf is not empty.
    :rtype: bool
    """
    if ds_InputRawDataObj.GetProtRatioDf().empty:
        print('Protein ratio is empty.')
        return False
    else:
        return True


def CheckMetadataNotEmpty(ds_InputRawDataObj):
    """
    Check metadata is not empty

    :param ds_InputRawDataObj:
    :type ds_InputRawDataObj: object
    :return: True if metadataDf is not empty.
    :rtype: bool
    """
    if ds_InputRawDataObj.GetMetadataDf().empty:
        print('Metadata is empty.')
        return False
    else:
        return True







import pandas as pd
from scipy.stats import mannwhitneyu


def RunWilconTest(filterRatioCsv, clusterCsv, clusterNum, fileName):
    """
    execute wilcoxon rank test to generate the genes' p-value file(to see if they pass the wilcon test) each subtype

    :param filterRatioCsv: the or the ratio table csv file that used to run wilcoxon rank test
    :param clusterCsv: the csv file that represent each sample's subtype
    :param clusterNum: the number of clusters
    :param fileName: the filename of the wilcoxon rank test results
    :return: None
    """
    if isinstance(filterRatioCsv, str):
        filterRatioDf = pd.read_csv(filterRatioCsv, index_col=0)
    else :
        filterRatioDf = filterRatioCsv.GetProtRatioDf().copy()

    if isinstance(clusterCsv, str):
        clusteringResultSer = pd.read_csv(clusterCsv, index_col=0, header=None)
        clusteringResultSer = clusteringResultSer.iloc[:, 0]
    else :
        clusteringResultSer = clusterCsv.clustResultObj.sampleClustSer

    # print(clusteringResultSer.tolist())

    for Num in range(clusterNum):
        clusterLi, nonclusterLi = _GetclusterLi(clusteringResultSer, Num + 1)
        clustfilterRatioDf = filterRatioDf.T[clusterLi]
        nonclusterfilterRatioDf = filterRatioDf.T[nonclusterLi]
        count = 0
        resultDf = pd.DataFrame()
        for geneStr in clustfilterRatioDf.index:
            count += 1
            stat, pval = mannwhitneyu(clustfilterRatioDf.T[geneStr], nonclusterfilterRatioDf.T[geneStr],
                                      use_continuity=False, alternative='greater')
            ser = pd.Series({'gene': geneStr, 'p-value': pval, 'u-stat':stat})
            resultDf = resultDf.append(ser, ignore_index=True)
        resultDf.sort_values(by='p-value', ascending=False, inplace=True)
        resultDf = resultDf[::-1]
        resultDf.to_csv(fileName + "_pval_" + str(Num + 1) + ".csv")
        print(resultDf)

    # wilcoxon()


def _GetclusterLi(clusteringResultSer, subtype):
    """
    get the current subtype's samples list and rest samples list, in order to run wilcoxon rank test

    :param clusteringResultSer: the series that represent each sample's subtype
    :param subtype:int, the current subtype
    :return: current subtype's samples list and rest samples list
    """
    clusterLi = []
    nonclusterLi = []
    for k in clusteringResultSer.index:
        if clusteringResultSer[k] == subtype:
            clusterLi.append(k)
        else:
            nonclusterLi.append(k)

    return clusterLi, nonclusterLi



import pandas as pd
import copy
from .CommonFunction import GenerateStat, PrintStat

def RemoveProtWithNa(rawDataObj):
    """
    Removed rows (features) which contain missing data

    :param rawDataObj: ds_InputRawData
    :type rawDataObj: object
    :return: protRatioDropNaDf
    :rtype: pd.DataFrame
    """

    print('Before removing NA')
    PrintStat(rawDataObj)

    protRatioDf = rawDataObj.GetProtRatioDf()

    protRatioDropNaDf = protRatioDf.dropna(axis=1)

    dropNaDataObj = copy.deepcopy(rawDataObj)
    dropNaDataObj.SetProtRatioDf(protRatioDf=protRatioDropNaDf)

    print('After removing NA')
    impDataObj = GenerateStat(dropNaDataObj)

    return dropNaDataObj


def AssignNaStrToNotAvailableCell(metadataDf):
    """
    Assign missing value of metadataDf to 'NA'
    
    Called by ReadFilesPath()

    :param metadataDf: 
    :type metadataDf: pd.DataFrame
    :return: newMetadataDf
    :rtype: pd.DataFrame
    """
    newMetadataDf = metadataDf.fillna('NA')
    return newMetadataDf


def RemoveContinuousFeature(metadataDf, metadataCategoryDf=pd.DataFrame()):
    """
    Drop continuous features (columns) of metadataDf that depends on metadataCategoryDf or check whether its label count is over half of length.
    
    Called by ReadFilesPath()

    :param metadataDf: 
    :type metadataDf: pd.DataFrame
    :param metadataCategoryDf: Consist of whether each feature is categorical.
    :return: newMetadataDf
    :rtype: pd.DataFrame
    """

    if not metadataCategoryDf.empty:
        '''Depends on metadataCategoryDf'''
        CategoricalColumnLi = metadataCategoryDf.transpose().index[(metadataCategoryDf.transpose().iloc[:, 0] == 1)]
        newMetadataDf = metadataDf[CategoricalColumnLi]
    else:
        '''check whether its label count is over half of length'''
        columnLength = len(metadataDf)
        CategoricalColumnLi = []
        for x in metadataDf.columns:
            if len(metadataDf[x].unique()) < columnLength / 3:
                CategoricalColumnLi.append(x)
        newMetadataDf = metadataDf[CategoricalColumnLi]

    print('Categorical features:')
    print(CategoricalColumnLi)

    return newMetadataDf




import math
from .CommonFunction import *

def LogRatioToRatio(rawDataObj):
    """

    :param rawDataObj: ds_InputRawData 
    :type rawDataObj: object
    :return: ds_InputRawData
    :rtype: object
    """

    print('Before changing to original ratio')
    PrintStat(rawDataObj)

    logRatioDf = rawDataObj.GetProtRatioDf()
    ratioDf = logRatioDf.applymap(lambda x: 2 ** x)
    rawDataObj.SetProtRatioDf(ratioDf)
    rawDataObj.numericalFormatStr = 'Original'

    print('After changing to original ratio')
    PrintStat(rawDataObj)

    return rawDataObj


def RatioToLogRatio(rawDataObj):
    """
    
    :param rawDataObj: ds_InputRawData 
    :type rawDataObj: object
    :return: ds_InputRawData
    :rtype: object
    """

    print('Before changing to log ratio')
    PrintStat(rawDataObj)

    ratioDf = rawDataObj.GetProtRatioDf()
    logRatio = ratioDf.applymap(lambda x: math.log2(x))
    rawDataObj.SetProtRatioDf(logRatio)
    rawDataObj.numericalFormatStr = 'Log'

    print('After changing to log ratio')
    PrintStat(rawDataObj)

    return rawDataObj



import numpy as np
import pandas as pd
from .CommonFunction import *
from sklearn.metrics.cluster import v_measure_score






def CalcFeatureVscore(rawDataObj, protClustObj, beta=1.0):
    """
    傳入inputRawData和protClust、beta及clustering結果，計算各個metadata feauture的Vscore並儲存至ds_protClust object中的featureVscoreDi dictionary

    :param rawDataObj: ds_inputRawData object
    :param protClustObj: ds_protClust object
    :param predLi: clutering result list，大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param beta: weights of the harmonic mean of homogeneity and completeness, default=1
    :return: ds_protClust object
    """

    metaColNameLi = rawDataObj.GetMetadataDf().columns.values
    metaDf = rawDataObj.GetMetadataDf().copy()
    predLi = protClustObj.clustResultObj.sampleClustSer.copy()
    #predLi=list(TurnClustToNumSer(rawDataObj.GetMetadataDf().index.values, protClustObj))
    protClustObj.clustScoringObj.featureVscoreDi = dict()
    for colNameStr in metaColNameLi:


        trueLi = TurnLabelToNum(metaDf[colNameStr])
        predLi = TurnLabelToNum(predLi)

        protClustObj.clustScoringObj.featureVscoreDi.update({colNameStr: v_measure_score(trueLi, predLi, beta)})




    return protClustObj



from .CommonFunction import *
from collections import Counter
from scipy.spatial import distance

def CalcBicScore(rawDataObj, protClustObj):
    """
    Calculate Bayesian information criterion score to estimate cluster number

    :param rawDataObj: ds_inputRawData object
    :type rawDataObj: object
    :param protClustObj: ds_protClust object
    :type protClustObj: object
    :return: ds_protClust object
    :rtype: object
    """

    rawDataArr = rawDataObj.GetProtRatioDf().values.copy()
    predLi = protClustObj.clustResultObj.sampleClustSer.values.copy() # cluster result
    clusterNum = len(np.unique(np.array(predLi))) # cluster number
    centerArr = GetClusterCenter(rawDataArr, predLi, clusterNum) # cluster center
    clusterSizeLi = list(Counter(predLi).values())
    # print(rawDataArr); print(predLi); print(clusterNum); print(centerArr); print(clusterSizeLi);print('\n\n\n')

    protClustObj.clustScoringObj.bicScore = \
        _Processing(rawDataArr=rawDataArr, centerArr=centerArr, predLi=predLi, clusterNum=clusterNum, clusterSizeLi=clusterSizeLi)

    return protClustObj


def _Processing(rawDataArr, centerArr, predLi, clusterNum, clusterSizeLi):
    """
    reference: https://stats.stackexchange.com/questions/90769/using-bic-to-estimate-the-number-of-k-in-kmeans

    :param rawDataArr: Protein ratio. Shape [data_num feature_num]
    :type rawDataArr: np.array
    :param centerArr: Center of the clusters, dimension [cluster_num, feature_num]
    :type centerArr np.array
    :param predLi: sampleClustSer's list [1, 2, 2, 3, 3, 4]
    :type predLi: list
    :param clusterNum: Number of cluster.
    :type clusterNum: int
    :param clusterSizeLi: List of each cluster size.
    :type clusterSizeLi: list
    :return: BIC score
    :rtype: float
    """

    rawDataSize, rawDataDim = rawDataArr.shape

    # compute variance for all clusters beforehand
    cl_var = (1.0 / (rawDataSize - clusterNum) / rawDataDim) * sum([sum(distance.cdist(rawDataArr[np.where(predLi == (i+1))],[centerArr[i]], 'euclidean') ** 2) for i in range(clusterNum)])
    const_term = 0.5 * clusterNum * np.log(rawDataSize) * (rawDataDim + 1)

    BIC = np.sum([clusterSizeLi[i] * np.log(clusterSizeLi[i]) -
                  clusterSizeLi[i] * np.log(rawDataSize) -
                  ((clusterSizeLi[i] * rawDataDim) / 2) * np.log(2 * np.pi * cl_var) -
                  ((clusterSizeLi[i] - 1) * rawDataDim / 2) for i in range(clusterNum)]) - const_term

    return BIC



from .CommonFunction import *

import numpy.linalg as LA


def CalcSeparationIndex(rawDataObj, protClustObj):
    """
    :param rawDataObj: ds_inputRawData object
    :param protClustObj: ds_protClust object
    :return: ds_protClust object
    """
    rawDataArr = rawDataObj.GetProtRatioDf().values.copy()

    predLi = protClustObj.clustResultObj.sampleClustSer.values.copy()

    clusterNum = len(np.unique(np.array(predLi)))
    centerArr = GetClusterCenter(rawDataArr, predLi, clusterNum)

    utilityArr = GetUtilityMatrix(rawDataArr, centerArr, 'euclidean')

    protClustObj.clustScoringObj.separationIndex = _Processing(rawDataArr, centerArr, utilityArr, m=2)

    return protClustObj


def _Processing(dataArr, centerArr, utilityArr, m=2):
    """
    Calculate the separation index
    :param: dataArr: np.array. Shape [data_num feature_num]
    :param: centerArr: np.array. Center of the clusters, dimension [cluster_num, feature_num]
    :param: utilityArr: np.array. Utility matrix, shape [cluster_num, data_num]
    :return: flaot partition index
    """

    sampleNum = dataArr.shape[0]  # (data_num, feature_num)
    clusterNum = centerArr.shape[0]  # (cluster_num, feature_num)

    compact = 0
    dist2Center = []
    for i in range(clusterNum):
        for j in range(i + 1, clusterNum):
            dist2Center.append(LA.norm(centerArr[i] - centerArr[j]))

        w = np.power(utilityArr[i], m)
        comp_clus = 0
        for k in range(sampleNum):
            comp_clus += w[k] * LA.norm(dataArr[k] - centerArr[i])
        compact += comp_clus

    dist2Center = np.array(dist2Center)
    S = (compact / (sampleNum * np.min(dist2Center)))

    return S



import matplotlib
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib_venn import venn2, venn3





def DrawNetworkOverlapGeneVenn(node1Csv, node2Csv, subtype, degreeCutOff=10):
    """
    draw venn diagram about the number of two subtypes' overlap genes

    :param node1Csv: str, cluster gene node csv path
    :param node2Csv: str, cluster gene node csv path
    :param subtype: number of clusters
    :param degreeCutOff: number of degree

    :return: None

    """
    node1Df = pd.read_csv (node1Csv, index_col=0)
    node2Df = pd.read_csv (node2Csv, index_col=0)
    count=0
    filterNode1Df=node1Df[node1Df.Degree>=degreeCutOff]
    filterNode2Df=node2Df[node2Df.Degree>=degreeCutOff]

    for name1 in filterNode1Df['Label']:
        for name2 in filterNode2Df['Label']:
            #if name1==name2 and name1!='Id':
            if name1 == name2 :
                count+=1
    fig, ax = plt.subplots (nrows=1, ncols=1)
    venn2 (subsets=(filterNode1Df.shape[0]-count,filterNode2Df.shape[0]-count,count ), set_labels=('ConKM', 'HC'),ax=ax)
    fig.savefig("vennfig_subtype"+str(subtype)+"_"+str(degreeCutOff),bbox_inches='tight')
    plt.close (fig)

    print("subtype"+str(subtype)+": ",filterNode1Df.shape[0],filterNode2Df.shape[0],count)




def DrawGoTermOverlapVenn(node1Csv, node2Csv, path, node1Name='ConKM', node2Name='HC',dpi=300):

    overlapLi=[]
    goTerm1Df = pd.read_csv(node1Csv, delimiter="\t")
    goTerm2Df = pd.read_csv(node2Csv, delimiter="\t")
    count = 0
    num = 0
    for goNameStr in goTerm1Df['geneSet']:
        num += 1

        for goName2Str in goTerm2Df['geneSet']:
            if goNameStr == goName2Str:
                count += 1
                overlapLi.append(goName2Str)



    # set font size
    font = {'family': 'sans',
            'weight': 'bold',
            'size': 22}
    matplotlib.rc('font', **font)
    fig, ax = plt.subplots(nrows=1, ncols=1,dpi=dpi)


    venn2(subsets=(len(goTerm1Df['geneSet']) - count, len(goTerm2Df['geneSet']) - count, count),
          set_labels=(node1Name, node2Name), ax=ax)
    fig.savefig(path, bbox_inches='tight')
    plt.close(fig)
    print(overlapLi)
    return overlapLi




def DrawOverlapOverlapVenn3(list1, list2, list3, path,node1='1', node2='3',node3='5',dpi=300):
    count1_2_3=0
    count1_2 = _CountTwoOverlap(list1,list2)
    count2_3 = _CountTwoOverlap(list2,list3)
    count1_3 = _CountTwoOverlap(list1,list3)
    overlapLi=[]
    for goNameStr in list1:


        for goName2Str in list2:
            for goName3Str in list3:

                if goNameStr == goName2Str == goName3Str:
                    count1_2_3 += 1
                    overlapLi.append(goName2Str)


    font = {'family': 'sans',
            'weight': 'bold',
            'size': 22}
    matplotlib.rc('font', **font)
    fig, ax = plt.subplots(nrows=1, ncols=1, dpi=dpi)

    venn3(subsets=(len(list1) - count1_2-count1_3 + count1_2_3, len(list2) - count2_3 - count1_2 + count1_2_3, count1_2-count1_2_3,
                   len(list3) - count2_3 - count1_3 + count1_2_3, count1_3-count1_2_3, count2_3-count1_2_3, count1_2_3),
          set_labels=(str(node1), str(node2), str(node3)), ax=ax)
    fig.savefig(path, bbox_inches='tight')
    plt.close(fig)


def _CountTwoOverlap(list1, list2):
    count=0
    for name1 in list1:

        for name2 in list2:
            if name1==name2:
                count+=1

    return count


import seaborn as sns
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import random
import math
from collections import Counter

from .CommonFunction import *


def DrawStackedBarForCombineFig(clustNumSer, currentFeatureSer, axes, color="Set1", edgecolor='white', width=0.5,
                                legendSize=10):
    """
    draw a stacked bar based on cluster and meta feature value

    :param clustNumSer: the number Series represents the samples' cluster. [1,0,0,2] means 4 samples in different 3 clusters
    :param currentFeatureSer: the Series includes each sample's feature value. ['white','black','white']
    :param axes: the postion of the bar plot in the combineFig
    :param color: the color set of different feature values
    :param edgecolor: the background color of the bar plot
    :param width: the width of each bar
    :param legendSize: the size of the legned
    :return: axes, the postion of the bar plot in the combineFig
    """


    clustXaxisLi=GetClustXaxisLi(clustNumSer)
    colors = sns.color_palette(color, len(set(currentFeatureSer)))

    featureValueMatrix = [[0 for x in range(len(set(clustNumSer)))] for y in range(len(set(currentFeatureSer)))]

    ##### for every predict cluster
    #for currentClust in range(len(set(clustNumSer))):
    clustCount=0
    for currentClust in set(clustNumSer):
        clusterFilter = (clustNumSer == currentClust)
        barClust = currentFeatureSer[clusterFilter]
        valueCount = 0
        # print(barClust.value_counts())
        ##### for every type in every meta data
        for value in (set(currentFeatureSer)):
            featureValueMatrix[valueCount][clustCount] = CountFeatureValueNum(barClust, value)
            valueCount += 1
            # else:
            #   Matrix[currentClust][valueCount]=0
    #      valueCount=valueCount+1
        clustCount+=1
    bars = [0 for x in range(len(set(clustNumSer)))]
    for i in range(len(set(currentFeatureSer))):
        axes.bar(clustXaxisLi, featureValueMatrix[i], bottom=bars, color=colors[i], edgecolor=edgecolor, width=width)
        bars = np.add(bars, featureValueMatrix[i]).tolist()
    axes.legend(set(currentFeatureSer), loc='lower left', prop={'size': legendSize}, bbox_to_anchor=(0.0, 1.1, 0.4, 0.3))
    axes.set_xlabel('cluster')
    axes.set_title(currentFeatureSer.name)

    return axes


def DrawStackedBar(protClustObj, currentFeatureSer, color="Set1", edgeColor='white', width=0.5,
                   legendSize=10, figSizeLi=[5, 4], path='stackedbar.png', dpi=72):
    """
    draw a stacked bar based on cluster and meta feature value

    :praram protClustObj: ds_protclustObj
    :param currentFeatureSer: the Series includes each sample's feature value. ['white','black','white']
    :param color: the color set of different feature values
    :param edgeColor: the background color of the bar plot
    :param width: the width of each bar
    :param figSizeLi: the width and high of the figure
    :param legendSize: the size of the legned
    :return: axes, the position of the bar plot in the combineFig
    """

    clustNumSer = protClustObj.clustResultObj.sampleClustSer
    clustXaxisLi = GetClustXaxisLi (clustNumSer)
    colors = sns.color_palette (color, len (set (currentFeatureSer)))
    fig, axes = plt.subplots (1, 1, figsize=(figSizeLi[0], figSizeLi[1]), dpi=dpi)
    featureValueMatrix = [[0 for x in range (len (set (clustNumSer)))] for y in range (len (set (currentFeatureSer)))]

    ##### for every predict cluster
    clustCount = 0
    for currentClust in set(clustNumSer):
        clusterFilter = (clustNumSer == currentClust)
        barClust = currentFeatureSer[clusterFilter]
        valueCount = 0
        # print(barClust.value_counts())
        ##### for every type in every meta data
        for value in (set (currentFeatureSer)):
            featureValueMatrix[valueCount][clustCount] = CountFeatureValueNum (barClust, value)
            valueCount += 1
            # else:
            #   Matrix[currentClust][valueCount]=0
    #      valueCount=valueCount+1
        clustCount+=1
    bars = [0 for x in range (len (set (clustNumSer)))]
    for i in range (len (set (currentFeatureSer))):
        axes.bar (clustXaxisLi, featureValueMatrix[i], bottom=bars, color=colors[i], edgecolor=edgeColor, width=width)
        bars = np.add (bars, featureValueMatrix[i]).tolist ()
    axes.legend (set (currentFeatureSer), loc='lower left', prop={'size': legendSize},
                 bbox_to_anchor=(0.0, 1.1, 0.4, 0.3))

    axes.set_title (currentFeatureSer.name)
    axes.set_xlabel('cluster')
    fig.savefig (path, bbox_inches="tight")

    return fig



from .CommonFunction import *

from scipy.special import comb

def CalcFeaturePairedFscore(rawDataObj, protClustObj):
    """
    Calculate the PairedFscore for each label and the clustering results

    :param rawDataObj: ds_inputRawData object
    :param protClustObj: ds_protClust object
    :return: ds_protClust object
    """

    metaColName = rawDataObj.GetMetadataDf().columns.values.copy()
    metaDf = rawDataObj.GetMetadataDf().copy()
    predLi = protClustObj.clustResultObj.sampleClustSer.values.copy()

    protClustObj.clustScoringObj.featurePairedFscoreDi = dict()
    for colName in metaColName:
        trueLi = TurnLabelToNum(metaDf[colName])
        predLi = TurnLabelToNum(predLi)

        protClustObj.clustScoringObj.featurePairedFscoreDi.update({colName: _GetSinglePairedFscore(trueLi, predLi)})
    return protClustObj


def _GetTnTpFnFp(trueLi, predLi):
    """
    Calculate the TP, TN, FP, FN pairs

    :param trueLi: feature class list,大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param predLi: clutering result list，大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :return: dict {'TP', 'FP', 'FN', 'FN'}
    """
    clustersArr = np.array(predLi)
    classesArr = np.array(trueLi)
    tp_plus_fp = comb(np.bincount(clustersArr), 2).sum()
    tp_plus_fn = comb(np.bincount(classesArr), 2).sum()
    A = np.c_[(clustersArr, classesArr)]
    tp = sum(comb(np.bincount(A[A[:, 0] == i, 1]), 2).sum()
             for i in set(clustersArr))
    fp = tp_plus_fp - tp
    fn = tp_plus_fn - tp
    tn = comb(len(A), 2) - tp - fp - fn
    return {'TP':tp,'FP':fp,'FN':fn,'TN':tn}

def _GetSinglePairedFscore(trueLi, predLi):
    """
    Calculate paried Fscore

    :param trueLi: feature class list,大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :param predLi: clutering result list，大小與sample數一樣，如 0, 0, 1, 1, 2, 0, 2 數字代表cluster index
    :return: float paired F score
    """
    pairNumDi = _GetTnTpFnFp(trueLi, predLi)
    return 2*pairNumDi['TP'] / (2*pairNumDi['TP'] + pairNumDi['FP'] + pairNumDi['FN'])





from .CommonFunction import *


def CalcFeatureRatio(rawDataObj, protClustObj):
    """
    :param rawDataObj: ds_inputRawData object
    :param protClustObj: ds_protClust object
    :return: ds_protClust object
    """
    metaDf = rawDataObj.GetMetadataDf().copy()
    predLi = protClustObj.clustResultObj.sampleClustSer.values.copy()

    bgRatioSer = pd.Series()  # feature -> featureValue -> featureValueRatio
    for featureName in metaDf.columns:
        bgRatioSer[featureName] = (metaDf[featureName].value_counts() / metaDf.shape[0])

    clustDi = {}

    for clusIdx in range(1, np.max(predLi) + 1):
        boolMask = (predLi == clusIdx)
        clustDi[clusIdx] = {}
        for featureName in metaDf.columns:
            featValueRatioSer = metaDf[featureName][boolMask].value_counts() / metaDf[boolMask].shape[0]
            # index: feature value, data: ratio in the cluster
            for featureValue in bgRatioSer[featureName].index:
                try:
                    featValueRatioSer[featureValue] = featValueRatioSer[featureValue] / bgRatioSer[featureName][
                        featureValue]
                except:
                    featValueRatioSer.at[featureValue] = 0
            clustDi[clusIdx][featureName] = featValueRatioSer
        clustDi[clusIdx] = pd.Series(clustDi[clusIdx])

    clustSer = pd.Series(clustDi)

    protClustObj.clustScoringObj.clustFeatureValueRatioSer = clustSer

    return protClustObj



import numpy as np
import pandas as pd
from sklearn import cluster, metrics, datasets
from .KMeans import *
from .Hierarchical import *
from .Consensus import *
from .GuassianModel import GaussianEMClust
from .SelfOrganizingMap import SelfOrganizingMapClust
from pylab import *

def Clustering(rawDataObj, protClustObj):
    """
    Top level of clustering to perform different cluster method including:\n
    KmeansClust_sklearn,  KmeansClust_nltk,  HierarchicalClust,
    ConsensusClust  ( KmeansClust_sklearn ),  ConsensusClust ( KmeansClust_nltk ),  ConsensusClust ( HierarchicalClust ),  ConsensusClust ( HierarchicalClust_Scipy ),
    GaussianEMClust,  SelfOrganizingMapClust

    :param rawDataObj:    use protein ratio dataframe to perform clustering
    :type rawDataObj: object
    :param protClustObj:  use cluster parameter saved in protClustObj.clustParamObj to perform clustering
    :type protClustObj: object
    :return protClustObj: protClustObj with saving utilityMatrixDf, clustSampleSer, sampleClustSer
    
    .. note::
        utilityMatrixDf : pandas dataframe (row is sample, column is cluster number, value is 1 if sample belongs cluster otherwise 0) \n
        clustSampleSer  : 2D pandas Series (1D is cluster number, 2D is sample) [[p1,p3,p4], [p2], [p5]] \n
        sampleClustSer  : pandas Series (size is sample amount; values are cluster ID) [1,2,1,2,2,3,2,1]
    
    """
    dataDf = rawDataObj.GetProtRatioDf()
    patientLi = list(dataDf.index)
    dataArr = dataDf.values
    methodStr = protClustObj.clustParamObj.GetParam('method')

    clusterResultLi = None
    CCCDF = None

    print('available parameter : {}'.format(list(protClustObj.clustParamObj.GetParamLi())))
    if methodStr == 'kmeansClust_sklearn':
        print('KmeansClust_sklearn starting ...')

        clusterResultLi = KmeansClust_sklearn(rawDataObj, protClustObj)
        print('KmeansClust_sklearn result :\n{}'.format(clusterResultLi))

    elif methodStr == 'kmeansClust_nltk':
        print('KmeansClust_nltk starting ...')

        clusterResultLi = KmeansClust_nltk(rawDataObj, protClustObj)
        print('KmeansClust_nltk result :\n{}'.format(clusterResultLi))

    elif methodStr == 'hierarchicalClust':
        print('HierarchicalClust starting ...')

        clusterResultLi = HierarchicalClust(rawDataObj, protClustObj)
        print('HierarchicalClust result :\n{}'.format(clusterResultLi))

    elif methodStr == 'hierarchicalClust_scipy':
        print('hierarchicalClust_scipy starting ...')

        clusterResultLi = HierarchicalClust_Scipy(rawDataObj,protClustObj)
        print('hierarchicalClust_scipy result :\n{}'.format(clusterResultLi))
        
    elif methodStr == 'consensusClust':
        print('ConsensusClust starting ...')

        ccMethodStr = protClustObj.clustParamObj.GetParam('ccMethod')
        rp = protClustObj.clustParamObj.GetParam('resample')
        minK = protClustObj.clustParamObj.GetParam('minK')
        maxK = protClustObj.clustParamObj.GetParam('maxK')
        resample_times = protClustObj.clustParamObj.GetParam('resample_times')
        if ccMethodStr == 'kmeansClust_sklearn':
            CC = ConsensusClust(cluster=cluster.KMeans, minK=minK, maxK=maxK, resample_times=resample_times,
                                resample_proportion=rp).fit(dataArr=dataArr)
            CC_predict = CC.Predict()
            print('ConsensusClust (KmeansClust_sklearn) result :\n{}'.format(CC_predict))

        elif ccMethodStr == 'kmeansClust_nltk':
            distance = protClustObj.clustParamObj.GetParam('ccDistance')
            CC = ConsensusClust(cluster=KMeansClusterer, minK=minK, maxK=maxK, resample_times=resample_times,
                                resample_proportion=rp, distance=distance).fit(dataArr=np.array(dataArr))
            CC_predict = CC.Predict()
            print('ConsensusClust (KmeansClust_nltk) result :\n{}'.format(CC_predict))

        elif ccMethodStr == 'hierarchicalClust':
            linkage = protClustObj.clustParamObj.GetParam('ccLinkage')
            distance = protClustObj.clustParamObj.GetParam('ccDistance')
            finallinkage = protClustObj.clustParamObj.GetParam('ccFinalLinkage') if protClustObj.clustParamObj.GetParam('ccFinalLinkage') else 'ward'
            finaldistance = protClustObj.clustParamObj.GetParam('ccFinalDistance') if protClustObj.clustParamObj.GetParam('ccFinalDistance') else 'euclidean'
            CC = ConsensusClust(cluster=AgglomerativeClustering, minK=minK, maxK=maxK, resample_times=resample_times,
                                resample_proportion=rp,
                                linkage=linkage,distance=distance,finallinkage=finallinkage,finaldistance=finaldistance).fit(dataArr=np.array(dataArr))
            CC_predict = CC.Predict()
            print('ConsensusClust (HierarchicalClust) result :\n{}'.format(CC_predict))

        elif ccMethodStr == 'hierarchicalClust_scipy':
            linkage = protClustObj.clustParamObj.GetParam('ccLinkage')
            distance = protClustObj.clustParamObj.GetParam('ccDistance')
            finallinkage = protClustObj.clustParamObj.GetParam('ccFinalLinkage') if protClustObj.clustParamObj.GetParam('ccFinalLinkage') else 'ward'
            finaldistance = protClustObj.clustParamObj.GetParam('ccFinalDistance') if protClustObj.clustParamObj.GetParam('ccFinalDistance') else 'euclidean'
            CC = ConsensusClust(cluster=sc.hierarchy.cut_tree, minK=minK, maxK=maxK, resample_times=resample_times,
                                resample_proportion=rp,
                                linkage=linkage, distance=distance, finallinkage=finallinkage, finaldistance=finaldistance).fit(dataArr=np.array(dataArr))
            CC_predict = CC.Predict()
            print('ConsensusClust (hierarchicalClust_scipy) result :\n{}'.format(CC_predict))

        else:
            CC_predict = None
            print('please provide available Clustering algorithm for ConsensusClust')  
        # X  = np.arange(len(CC.CDF[0]))
        # for i in CC.CDF:
        #     plot(CC.CDF[i])
        # show()
        CCCDF = CC.CDF
        clusterResultLi = CC_predict

    elif methodStr == 'gaussianEMClust':
        print('GaussianEMClust starting ...')

        clusterResultLi = GaussianEMClust(rawDataObj, protClustObj)
        print('GaussianMixture Clustering result :\n{}'.format(clusterResultLi))

    elif methodStr == 'selfOrganizingMapClust':
        print('SelfOrganizingMapClust starting')

        clusterResultLi = SelfOrganizingMapClust(rawDataObj, protClustObj)
        print('SelfOrganizingMapClust result :\n{}'.format(clusterResultLi))

    else:
        print('please provide available Clustering algorithm')
    clusterResultLi = [clust + 1 for clust in clusterResultLi]
    protClustObj.clustResultObj.utilityMatrixDf = _GetUtilityMatrixDf(clusterResultLi, patientLi)
    protClustObj.clustResultObj.clustSampleSer = _GetClustSampleSer(clusterResultLi, patientLi)
    protClustObj.clustResultObj.sampleClustSer = pd.Series(clusterResultLi, patientLi)
    protClustObj.clustResultObj.clustNum = max(clusterResultLi)
    protClustObj.clustResultObj.cdfLi = CCCDF
    return protClustObj

def ReadClustSeries(clusterResultLi,rawDataObj, protClustObj):
    """
    Saving a clustered result into data structure without performing clustering method

    :param clusterResultLi: the clustered result
    :type clusterResultLi: list
    :param rawDataObj: data structure to be saved
    :type rawDataObj: object
    :param protClustObj: data structure to be saved
    :type protClustObj: object

    """
    print(clusterResultLi)
    clusterResultLi = [clust + 1 for clust in clusterResultLi]
    dataDf = rawDataObj.GetProtRatioDf()
    patientLi = list(dataDf.index)
    protClustObj.clustResultObj.utilityMatrixDf = _GetUtilityMatrixDf(clusterResultLi, patientLi)
    protClustObj.clustResultObj.clustSampleSer = _GetClustSampleSer(clusterResultLi, patientLi)
    protClustObj.clustResultObj.sampleClustSer = pd.Series(clusterResultLi, patientLi)
    protClustObj.clustResultObj.clustNum = max(clusterResultLi)
    return protClustObj

def _GetUtilityMatrixDf(clusterResultLi, patientLi):
    """
    produce Utility Matrix Dataframe

    :param clusterResultLi: 
    :type clusterResultLi: list
    :param patientLi:
    :type patientLi: list
    :returns:  resultDf (pd.DataFrame())

    """
    resultDf = pd.DataFrame()
    clustNum = len(set(clusterResultLi))
    resultDf['sample'] = patientLi
    for i in range(1, clustNum + 1):
        clust = []
        for j in clusterResultLi:
            clust.append(1) if i == j else clust.append(0)
        resultDf['cluster' + str(i)] = clust
    resultDf = resultDf.set_index('sample')
    return resultDf


def _GetClustSampleSer(clusterResultLi, patientLi):
    """
    produce Cluster Sample Series
    :param clusterResultLi: 
    :type clusterResultLi: list
    :param patientLi:
    :type patientLi: list
    :returns: resultDf (pd.Series())
    """
    clustNum = len(set(clusterResultLi))
    index = [i for i in range(1, clustNum + 1)]
    data = [[] for i in range(clustNum)]
    for i in range(len(clusterResultLi)):
        data[clusterResultLi[i] - 1].append(patientLi[i])
    resultSer = pd.Series(data, index)
    return resultSer



# -*- coding: utf-8 -*-


from .CommonFunction import *
from .StackedBar import *
from .DivergeBar import *
from .ScatterPlot import *


def DrawCombinedFig(protClustObj, currentFeatureSer, compareFeatureSer, figTitleStr='Combined Figure',
                    figSizeLi=[15, 4], dpi=100, wSpace=0.7
                    , pathStr='combinFig.png', colorsStr='Set1', orderLi=['stacked', 'scatter', 'diverge']):
    """
    draw a combined figure included stacked, scatter, and diverge figure of clustering result

    :param protClustObj: ds_protClust object
    :param figSizeLi: the width and high of the figure
    :param currentFeatureSer: the Series includes each sample's feature value. ['white','black','white']
    :param compareFeatureSer: the Series includes you want to compare with currentFeatureSer
    :param pathStr: string, the name of the figure
    :param dpi: int, figure dpi
    :param wSpace: the width space between different figures
    :param orderLi: the string list include what plot you want to show in a CombineFig and the position of each plots
    :return: fig, the combination of subplots
    """

    clustNumSer = protClustObj.clustResultObj.sampleClustSer

    subplotNum = len(orderLi)
    fig, axes = plt.subplots(1, subplotNum, figsize=(figSizeLi[0] * subplotNum, figSizeLi[1]), gridspec_kw={'wspace': wSpace},
                             dpi=dpi)

    ######setting the order of plots######

    for orderNum in range(subplotNum):
        try:
            if orderLi[orderNum] != 'stacked' and orderLi[orderNum] != 'scatter' and orderLi[orderNum] != 'diverge':
                raise RuntimeError('use wrong plot name :' + orderLi[orderNum])
            elif len(set(orderLi)) < len(orderLi):
                raise RuntimeError('can not draw same plot twice')
            elif orderLi[orderNum] == 'stacked':

                axes[orderNum] = DrawStackedBarForCombineFig(clustNumSer, currentFeatureSer, axes[orderNum], color=colorsStr,
                                                          edgecolor='white',
                                                          width=0.5)

            elif orderLi[orderNum] == 'scatter':

                axes[orderNum] = DrawScatterPlotForCombineFig(clustNumSer, currentFeatureSer, compareFeatureSer,
                                                           axes[orderNum], color=colorsStr)

            elif orderLi[orderNum] == 'diverge':
                # scoringLi, scoreCenter = __DetermineScoreMethod(scoreMethod, protClustObj, clustNumber, currentFeatureName)
                axes[orderNum] = DrawDivergeBarForCombineFig(protClustObj, currentFeatureSer, axes[orderNum])

        except RuntimeError as e:
            print(e)

    #####use bbox_inches="tight" to make sure the completeness of legends in subplots#####
    fig.suptitle(figTitleStr, fontsize=16, x=0.5, y=1.5)
    fig.savefig(pathStr, bbox_inches="tight")

    return fig


def _GetEntropyFromProtClust(protClustObj, clustNum, currentFeatureName):
    """
    get entropy series from protClustObj in order to draw diverge bar

    :param protClustObj: ds_protClust object
    :param clustNum:int, the number of clusters
    :param currentFeatureName: string ,the feature name
    :return: entropy list
    """
    scoreLi = [0.0 for x in range(clustNum)]
    for currentClust in range(clustNum):
        scoreLi[currentClust] = protClustObj.clustScoringObj.clustFeatureEntropySer[currentClust][currentFeatureName] / \
                                protClustObj.clustScoringObj.featureEntropyDi[currentFeatureName]

    return scoreLi



from .CommonFunction import *
import scipy.stats as ss


def CalcFeatureEntropy(rawDataObj, protClustObj):
    """
    Calculate the entropy of each feature in metadata and save it in protClustObj.clustScoringObj.featureEntropyDi.\n
    No clustering result is required.

    :param rawDataObj:
    :param protClustObj:
    :return:
    """
    metaDf = rawDataObj.GetMetadataDf().copy()
    featureEntropyDi = {}

    for featureName in metaDf.columns:
        featureEntropyDi[featureName] = _ComputeEntropy(metaDf[featureName])

    protClustObj.clustScoringObj.featureEntropyDi = featureEntropyDi
    return protClustObj


def CalcClustFeatureEntropy(rawDataObj, protClustObj):
    """
    Calculate the feature entropy of each class and save it in protClustObj.clustScoringObj.clustFeatureEntropySer

    :param rawDataObj: ds_inputRawData object
    :param protClustObj: ds_protClust object
    :return: ds_protClust object
    """
    metaDf = rawDataObj.GetMetadataDf()
    predLi = protClustObj.clustResultObj.sampleClustSer.values.copy()

    clustSer = pd.Series()
    for clustIdx in range(1, np.max(predLi) + 1):
        clustDf = metaDf[predLi == clustIdx]
        clustSer.at[clustIdx] = pd.Series()
        for featureName in metaDf.columns:
            clustSer[clustIdx].at[featureName] = _ComputeEntropy(clustDf[featureName])

    protClustObj.clustScoringObj.clustFeatureEntropySer = clustSer
    return protClustObj


def _ComputeEntropy(featureSer):
    """
    compute entropy of a certain feature for a certain cluster

    :param featureSer {pd.Series}: the elements are feature values in a cluster. ex: {'patient name1': 'feature value 1', ...}
    :return: entropy {float}
    """
    countArr = featureSer.value_counts().values
    countArr = countArr / np.sum(countArr)
    return ss.entropy(countArr, base=2)



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm


def DrawSilhouettePlot(rawDataObj, protClustObj, path=None):
    clusterResult = protClustObj.clustResultObj.sampleClustSer

    silhScore = protClustObj.clustScoringObj.silhScore
    sampleSilhScore = protClustObj.clustScoringObj.sampleSilhScoreSer

    clusterNum = np.max(clusterResult) + 1
    fig, ax1 = plt.subplots()
    ax1.set_xlim([np.min(sampleSilhScore.values) - 0.05, np.max(sampleSilhScore.values) + 0.05])
    ax1.set_ylim([0, len(clusterResult) + (clusterNum + 1) * 10])

    y_lower = 10
    for i in range(1, clusterNum):
        clusterSilhouetteScoreArr = sampleSilhScore[clusterResult == i].values
        clusterSilhouetteScoreArr.sort()
        clusterSize = clusterSilhouetteScoreArr.shape[0]
        y_upper = y_lower + clusterSize
        color = cm.nipy_spectral(float(i) / clusterNum)

        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                          0, clusterSilhouetteScoreArr,
                          facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * clusterSize, str(i))

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples
    ax1.set_title("The Silhouette plot for the various clusters.")
    ax1.set_xlabel("Silhouette Score")
    ax1.set_ylabel("Cluster Label")

    # The vertical line for average silhouette score of all the values
    ax1.axvline(x=silhScore, color="red", linestyle="--")
    ax1.set_yticks([])  # Clear the yaxis labels / ticks
    ax1.set_xticks(np.arange(0, np.max(sampleSilhScore.values) + 0.05, 0.1))
    if path is None:
        plt.show()
    else:
        plt.savefig(path)



import numpy as np
import matplotlib.pyplot as plt


def SelfOrganizingMapClust(rawDataObj, protClustObj):
    """

    :param rawDataObj: ds_InputRawData
    :type rawDataObj: object 
    :param protClustObj: ds_ProtClust
    :type protClustObj: object
    :returns: predLi

    .. note::
        The parameters in protClustObj.clustParamObj.__kwargs must include 'dimension', 'learnRate', 'sigma', 'epochNum'\n
        Optional parameter: 'plotError'\n
        'dimension' can be a tuple for 2D maps, or an integer for 1-D maps.\n
        If 'plotError (bool)' is in the parameter list, the error history will be plotted.
    """
    dim = protClustObj.clustParamObj.GetParam('dimension')
    alpha = protClustObj.clustParamObj.GetParam('learnRate')
    sigma = protClustObj.clustParamObj.GetParam('sigma')
    epochNum = protClustObj.clustParamObj.GetParam('epochNum')

    rawDataArr = rawDataObj.GetProtRatioDf().values
    if isinstance(dim, int):
        som = SOM1D(dim)
    else:
        som = SOM2D(*dim)

    som.InitializeParams(rawDataArr)
    som.Fit(rawDataArr, alpha, sigma, epochNum)
    predLi = som.GetClustLabels(rawDataArr)
    errorHistory = som.GetErrorHistory()


    print('Error History')
    print(errorHistory)

    try:
        if protClustObj.clustParamObj.GetParam('plotError') == True:
            try:
                PlotErrorHistory(errorHistory, filename=protClustObj.clustParamObj.GetParam('filename'))
            except:
                PlotErrorHistory(errorHistory)
    except:
        pass

    return predLi


def PlotErrorHistory(errorLi, filename='SOM Error History.png'):
    """
    Plot the Error History

    :param errorLi:
    :type errorLi: list
    :return: None
    """
    plt.plot(errorLi)
    plt.xlabel('Epoch')
    plt.ylabel('Error')
    plt.title('SOM Error History')
    plt.savefig(filename)


class SOM2D(object):
    def __init__(self, xdim, ydim):
        self.__xdim = xdim  # number of nodes in x-dimension
        self.__ydim = ydim  # number of noder in y-dimension

        self.__nodeArr = np.array([])  # not initialized
        self.__alphaStart = 0
        self.__sigmaStart = 0
        self.__epochNum = 0
        self.__errorHistory = []
        self.__interval = 0  # interval of epochs to save the error history

    def InitializeParams(self, rawDataArr):
        """
        Initialize the nodes in self.__map with normal distribution

        :param rawDataArr: shape = (sampleNum, featureNum)
        :type rawDataArr: np.array
        """
        self.__nodeArr = np.random.normal(np.mean(rawDataArr), np.std(rawDataArr),
                                          size=(self.__xdim, self.__ydim, rawDataArr.shape[1]))

    def Fit(self, rawDataArr, alpha, sigma, epochNum, interval=1000):
        """
        Train the SOM on the given data for several iterations

        :param rawDataArr:  data to train on
        :type rawDataArr: np.array
        :param epochNum: number of iterations to train; if 0, epochs=len(data) and every data point is used once
        :type epochNum: int
        :param interval: interval of epochs to use for saving training errors
        :type interval: int
        """
        self.__interval = interval
        self.__alphaStart = alpha
        self.__sigmaStart = sigma
        self.__epochNum = epochNum

        for epoch in range(self.__epochNum):
            self.__TrainPerEpoch(rawDataArr, epoch)

    def GetClustLabels(self, rawDataArr):
        """
        Treat each unit in the nodeArr as a cluster. The best match unit is the cluster label for each data sample.
        The cluster index will be adjusted since some units are never used.

        :param rawDataArr:
        :type rawDataArr: np.array
        :return: clustLabels
        """
        clustLabels = []

        for dataSample in rawDataArr:
            x, y = self.__FindBestMatch(dataSample)
            clustLabels.append(x * self.__xdim + y)

        usedLabels = np.unique(clustLabels)
        transform = {usedLabels[i]: i for i in range(len(usedLabels))}
        clustLabels = [transform[clustLabels[i]] for i in range(len(clustLabels))]

        return np.array(clustLabels)

    def GetErrorHistory(self):
        return self.__errorHistory

    def __TrainPerEpoch(self, rawDataArr, currentEpoch):
        '''
        Update the SOM for one epoch. The training parameters are updated once in an epoch.

        :param rawDataArr: 
        :type rawDataArr: np.array
        :param currentEpoch: 
        :type currentEpoch: int
        '''
        alpha = self.__alphaStart / (1 + (currentEpoch / 0.5) ** 4)
        sigma = self.__sigmaStart / (1 + (currentEpoch / 0.5) ** 4)

        idxArr = np.random.choice(np.arange(rawDataArr.shape[0]), rawDataArr.shape[0], replace=False)

        # Traversing all data points in one epoch
        for dataidx in range(idxArr.shape[0]):
            self.__LearnOneSample(rawDataArr[dataidx], alpha, sigma)
        self.__errorHistory.append(self.__GetError(rawDataArr))

    def __GetError(self, rawDataArr):
        """
        get current error on all data samples

        :param rawDataArr: 
        :type rawDataArr: np.array
        """
        error = 0
        for dataSample in rawDataArr:
            error += np.min(np.sum((self.__nodeArr - dataSample) ** 2, axis=2))
        return error

    def __LearnOneSample(self, dataSample, alpha, sigma):
        '''
        Update the nodeArr based on one sample

        :param dataSample: one-dimension of a patient sample with different proteins
        :type dataSample: np.array
        :param alpha: current alpha
        :type alpha: float
        :param sigma: current sigma
        :type sigma: sigma
        :return: None
        '''
        xIdx, yIdx = self.__FindBestMatch(dataSample)
        distArr = self.__GetDist2BestMatch(xIdx, yIdx)
        h = np.exp(-(distArr / sigma) ** 2).reshape(self.__xdim, self.__ydim, 1)
        self.__nodeArr = self.__nodeArr - h * alpha * (self.__nodeArr - dataSample)

    def __FindBestMatch(self, dataSample):
        """
        Return the x-index and y-index for the best matching unit in the nodeArr

        :param dataSample: one-dimension of a patient sample with different proteins
        :type dataSample: np.array
        :return: xIdx {int}, y{int}
        """
        assert (dataSample.shape[0] == self.__nodeArr.shape[2])
        idx = np.argmin(np.sum((self.__nodeArr - dataSample) ** 2, axis=2))

        xIdx = int(idx / self.__nodeArr.shape[1])
        yIdx = idx % self.__nodeArr.shape[1]

        return xIdx, yIdx

    def __GetDist2BestMatch(self, xIdx, yIdx):
        """
        Return the distance matrix to the best match unit. The distance measure is Manhattan Distance

        :param xIdx: the x-index of the best match unit
        :type xIdx: int
        :param yIdx: the y-index of the best match unit
        :type yIdx: int
        """
        idxArr = np.array(list(np.ndindex(self.__xdim, self.__ydim))).reshape(self.__xdim, self.__ydim, 2)
        distArr = idxArr - np.array([xIdx, yIdx])
        distArr = np.sum(np.abs(distArr), axis=2)
        return distArr


class SOM1D(object):
    def __init__(self, dim):
        self.__dim = dim

        self.__nodeArr = np.array([])  # not initialized
        self.__alphaStart = 0
        self.__sigmaStart = 0
        self.__epochNum = 0
        self.__errorHistory = []
        self.__interval = 0  # interval of epochs to save the error history

    def InitializeParams(self, rawDataArr):
        """
        Initialize the nodes in self.__map with normal distribution

        :param rawDataArr: shape = (sampleNum, featureNum)
        :type rawDataArr: np.array
        """
        self.__nodeArr = np.random.normal(np.mean(rawDataArr), np.std(rawDataArr),
                                          size=(self.__dim, rawDataArr.shape[1]))

    def Fit(self, rawDataArr, alpha, sigma, epochNum, interval=1000):
        ''' Train the SOM on the given data for several iterations

        :param rawDataArr: data to train on
        :type rawDataArr: np.array
        :param epochNum: number of iterations to train; if 0, epochs=len(data) and every data point is used once
        :type epochNum: int
        :param interval: interval of epochs to use for saving training errors
        :type interval: int
        '''
        self.__interval = interval
        self.__alphaStart = alpha
        self.__sigmaStart = sigma
        self.__epochNum = epochNum

        for epoch in range(self.__epochNum):
            self.__TrainPerEpoch(rawDataArr, epoch)

    def GetClustLabels(self, rawDataArr):
        """
        Treat each unit in the nodeArr as a cluster. The best match unit is the cluster label for each data sample.
        The cluster index will be adjusted since some units are never used.

        :param rawDataArr:
        :type rawDataArr: np.array
        :return: clustLabels
        """
        clustLabels = []

        for dataSample in rawDataArr:
            idx = self.__FindBestMatch(dataSample)
            clustLabels.append(idx)

        usedLabels = np.unique(clustLabels)
        transform = {usedLabels[i]: i for i in range(len(usedLabels))}
        clustLabels = [transform[clustLabels[i]] for i in range(len(clustLabels))]

        return np.array(clustLabels)

    def GetErrorHistory(self):
        return self.__errorHistory

    def __TrainPerEpoch(self, rawDataArr, currentEpoch):
        '''
        Update the SOM for one epoch. The training parameters are updated once in an epoch.

        :param rawDataArr:
        :type rawDataArr: np.array
        :param currentEpoch: 
        :type currentEpoch: int
        '''
        alpha = self.__alphaStart / (1 + (currentEpoch / 0.5) ** 4)
        sigma = self.__sigmaStart / (1 + (currentEpoch / 0.5) ** 4)

        idxArr = np.random.choice(np.arange(rawDataArr.shape[0]), rawDataArr.shape[0], replace=False)

        # Traversing all data points in one epoch
        for dataidx in range(idxArr.shape[0]):
            self.__LearnOneSample(rawDataArr[dataidx], alpha, sigma)
        self.__errorHistory.append(self.__GetError(rawDataArr))

    def __GetError(self, rawDataArr):
        """
        get current error on all data samples

        :param rawDataArr:
        :type rawDataArr: np.array
        :return: error
        """
        error = 0
        for dataSample in rawDataArr:
            error += np.min(np.sum((self.__nodeArr - dataSample) ** 2, axis=1))
        return error

    def __LearnOneSample(self, dataSample, alpha, sigma):
        """
        Update the nodeArr based on one sample

        :param dataSample: one-dimension of a patient sample with different proteins
        :type dataSample: np.array
        :param alpha: current alpha
        :type alpha: float
        :param sigma: current sigma
        :type sigma: sigma
        :return: None
        """
        idx = self.__FindBestMatch(dataSample)
        distArr = self.__GetDist2BestMatch(idx)
        h = np.exp(-(distArr / sigma) ** 2).reshape(-1, 1)
        self.__nodeArr = self.__nodeArr - h * alpha * (self.__nodeArr - dataSample)

    def __FindBestMatch(self, dataSample):
        """
        Return the x-index and y-index for the best matching unit in the nodeArr

        :param dataSample: one-dimension of a patient sample with different proteins
        :type dataSample: np.array
        :return: xIdx {int}, y{int}
        """
        assert (dataSample.shape[0] == self.__nodeArr.shape[1])
        idx = np.argmin(np.sum((self.__nodeArr - dataSample) ** 2, axis=1))

        return idx

    def __GetDist2BestMatch(self, idx):
        """
        Return the distance matrix to the best match unit. The distance measure is Manhattan Distance

        :param idx: the index of the best match unit
        :type idx: int
        :return: distArr
        """
        idxArr = np.arange(self.__dim)
        distArr = np.abs(idxArr - idx)
        return distArr




import pandas as pd
from scipy.spatial.distance import pdist, squareform
from scipy.spatial import distance


def GetDistance(dist):
	distanceDict = {
					'euclidean':distance.euclidean,
					'cosine':distance.cosine,
					'manhattan':distance.cityblock,
					'pcc':distance.correlation
					}
	return distanceDict[dist]

def GeneratePairwiseDistMat(dataDf=pd.DataFrame(), distanceStr='correlation'):
	"""
	2020/01/10
	Generate pairwise distances between observations for "Hierarchical.py".
	Support Hierarchical.hclust()

	:param dataDf:  (DataFrame) proteinRatio
	:param distanceLi: (String)
	:return:
	"""
	distanceArr = squareform(pdist(X=dataDf, metric=distanceStr))
	return distanceArr



import pandas as pd


def Characterize(protClustObj, cutoffUpper, cutoffLower):
    """
    Compare feature odds ratio between clusters and background(Metadata)

    :param protClustObj: ds_ProtClust
    :type protClustObj: object
    :param cutoffUpper: to get the feature which odds ratio is higher than ceiling threshold
    :type cutoffUpper: int
    :param cutoffLower: to get the feature which odds ratio is lower than bottom threshold
    :type cutoffLower: int
    :returns: protClustObj (ds_ProtClust) with new characterRatioDf, characterFeatureDf
    """
    assert cutoffUpper > 0, 'threshold should >0'
    up_threshold = cutoffUpper
    assert cutoffLower > 0, 'threshold should >0'
    down_threshold = cutoffLower

    if protClustObj.clustScoringObj.clustFeatureValueRatioSer.empty:
        print('please run scoring function first')
    else:
        count = 1
        outputRatioDi = {}
        outputFeatureDi = {}
        clustSer = protClustObj.clustResultObj.clustSampleSer.values # use to get the amount of sample in each cluster
        for cluster in protClustObj.clustScoringObj.clustFeatureValueRatioSer:
            name = "cluster {} ( sample amount :{} )".format(count,len(clustSer[count-1]))
            characterRatioLi = []
            characterFeatureLi = []
            labelNameLi = []
            for label in cluster:
                temp1 = label[label >= up_threshold]
                temp2 = label[label > 0][label <= down_threshold]
                ratio = ';'.join([str(i) for i in temp1] + [str(i) for i in temp2])
                feature = ';'.join([str(i) for i in temp1.index] + [str(i) for i in temp2.index])
                if len(ratio) > 0:
                    characterRatioLi.append(ratio)
                    characterFeatureLi.append(feature)
                else:
                    characterRatioLi.append('NA')
                    characterFeatureLi.append('NA')
                labelNameLi.append(label.name)
            outputRatioDi.update({name: characterRatioLi})
            outputFeatureDi.update({name: characterFeatureLi})
            count += 1
        outcomeRatioDf = pd.DataFrame.from_dict(outputRatioDi)
        outcomeFeatureDf = pd.DataFrame.from_dict(outputFeatureDi)
        outcomeRatioDf.index = labelNameLi
        outcomeFeatureDf.index = labelNameLi
        outcomeRatioDf = outcomeRatioDf.T
        outcomeFeatureDf = outcomeFeatureDf.T
    PrintDescription(outcomeRatioDf, outcomeFeatureDf, cutoffUpper, cutoffLower)

    protClustObj.clustResultObj.characterRatioDf = protClustObj.clustResultObj.characterRatioDf.append(outcomeRatioDf)
    protClustObj.clustResultObj.characterFeatureDf = protClustObj.clustResultObj.characterFeatureDf.append(
        outcomeFeatureDf)
    return protClustObj


def PrintDescription(ratioDf, featureDf, cutoffUpper, cutoffLower):
    """
    log the feature filtered by threshold for each cluster

    :param ratioDf: odds ratio of each feature in each cluster
    :type ratioDf: pd.DataFrame
    :param feature: each cluster remained feature after filtering
    :type feature: pd.DataFrame
    :param cutoffUpper: to get the feature which odds ratio is higher than ceiling threshold
    :type cutoffUpper: int
    :param cutoffLower: to get the feature which odds ratio is lower than bottom threshold
    :type cutoffLower: int
    """
    if ratioDf.empty or featureDf.empty:
        print('run Characterization first')
    else:
        print("Characterization (ratio) ...")
        print(ratioDf)
        print("Characterization (feature) ...")
        print(featureDf)
        print("\n Description of each cluster (odds ratio >= {})\n".format(cutoffUpper))
        for i in ratioDf.index:
            print(i)
            for j in ratioDf.columns:
                ratioLi = ratioDf.loc[i, j].split(';')
                featureLi = featureDf.loc[i, j].split(';')
                for k in range(len(featureLi)):
                    if featureLi[k] != 'NA' and float(ratioLi[k]) > cutoffUpper:
                        print("{0:25} : {1:20}, odds ratio: {2:5.2f}".format(j, featureLi[k], float(ratioLi[k])))
            print('\n')

        print("\n Description of each cluster (odds ratio <= {})\n".format(cutoffLower))
        for i in ratioDf.index:
            print(i)
            for j in ratioDf.columns:
                ratioLi = ratioDf.loc[i, j].split(';')
                featureLi = featureDf.loc[i, j].split(';')
                for k in range(len(featureLi)):
                    if featureLi[k] != 'NA' and float(ratioLi[k]) < cutoffLower:
                        print("{0:25} : {1:20}, odds ratio: {2:5.2f}".format(j, featureLi[k], float(ratioLi[k])))
            print('\n')
